#ifndef _SILECS_DESIGN_H_
#define _SILECS_DESIGN_H_

#include <silecs-communication/interface/equipment/SilecsDevice.h>
#include <silecs-communication/interface/communication/SilecsConnection.h>
#include <silecs-communication/interface/equipment/SilecsParamConfig.h>

#include <string>
#include <vector>

namespace Silecs
{

class XMLParser;

class SilecsDesign
{
public:
    SilecsDesign(const XMLParser& xmlParser, const std::string& className, Connection& plcConnection, const SilecsParamConfig& paramConfig);

    const std::string& getName() const;

    /*!
     * \return List of the related Device instances for the PLC.
     */
    std::vector<std::string> getDeviceList() const;

    /*!
     * \brief Returns one device instance of that PLC by label or index
     * \param deviceName label or index of the device
     * \return Reference of the device object
     */
    Device* getDevice(const std::string& deviceName);

    const std::vector<Device>& getDevices() const;

    /*!
     * \brief Provides the list of Blocks defined in the corresponding Class/Version.
     * At least one PLC must be attached (connected) to the Cluster, otherwise getBlockList()
     * returns an empty string.
     * \param select access-type of the blocks (Input, Output, InOut)
     * \return String list (space separated value) of Blocks name that have selected access-type
     */
    std::string getBlockList(AccessType accessType) const;

    /*!
     * \brief Copies input buffer in output buffer
     * The method copies the input buffer in the output buffer for each READ-WRITE register in the PLC for the block passed as parameter
     * \param blockName name of the block for which the copy must be performed
     */
    void copyInToOut(const std::string& blockName);

    void updateLocalData();

    int receiveForAllDevices(const std::string& blockName);
    int sendForAllDevices(const std::string& blockName);

private:

    /*!
     * \fn getBlock
     * \brief returns instance of the first found block of the first device
     */
    const std::unique_ptr<Block>& getOneDeviceBlock(const std::string& blockName) const;

    int recvAllBlockMode(const std::string& blockName);
    int sendAllBlockMode(const std::string& blockName);

    /*!
     * \brief Used for detemining the required block buffer size in case of BLOCK_MODE. In BLOCK_MODE a single block contains a specific register for all the devices,
     * hence when calculating the required buffer size, we need to only reserve the space for the devices which actually contain this block.
     * \param blockName name of the block which should be used to count the devices
     */
    unsigned int getNrDevicesWithBlock(const std::string& blockName);

    std::vector<Device> devices_;
    std::string className_;

    Connection& plcConnection_;

    SilecsParamConfig paramConfig_;
};

} // namespace Silecs

#endif // _SILECS_DESIGN_H_